# -*- coding: utf-8 -*-

import logging

import numpy as np
import progressbar

from jack.core import QASetting
from jack.io.load import loaders

logger = logging.getLogger(__name__)


def evaluate(reader, dataset, batch_size):
    triples = {tuple(q.question.split()) for q, a in dataset if a[0].text == "True"}
    entity_set = reader.shared_resources.entity_to_index

    conf = reader.shared_resources.config
    all_triples = set()
    if conf.get('train'):
        all_triples.update(tuple(qa.question.split()) for qa, a in loaders[conf['loader']](conf['train'])
                           if a[0].text == "True")
    if conf.get('dev'):
        all_triples.update(tuple(qa.question.split()) for qa, a in loaders[conf['loader']](conf['dev'])
                           if a[0].text == "True")
    if conf.get('test'):
        all_triples.update(tuple(qa.question.split()) for qa, a in loaders[conf['loader']](conf['test'])
                           if a[0].text == "True")

    def scoring_function(triples):
        scores = []
        for i in range(0, len(triples), batch_size):
            batch_qas = [
                QASetting(" ".join(triples[k]))
                for k in range(i, min(len(triples), (i + batch_size)))]
            if batch_qas:
                for a in reader(batch_qas):
                    scores.append(a[0].score)
        return scores

    ranks, filtered_ranks = compute_ranks(scoring_function, triples, entity_set, all_triples)
    results = dict()
    results['Unfiltered Results'] = ranking_summary(ranks)
    results['Filtered Results'] = ranking_summary(filtered_ranks)
    return results


def compute_ranks(scoring_function, triples, entity_set, true_triples=None):
    subject_ranks, object_ranks = [], []
    subject_ranks_filtered, object_ranks_filtered = [], []

    bar = progressbar.ProgressBar(
        max_value=len(triples),
        widgets=[' [', progressbar.Timer(), '] ', progressbar.Bar(), ' (', progressbar.ETA(), ') '])
    for s, p, o in bar(triples):
        subject_triples = [(s, p, o)] + [(x, p, o) for x in entity_set if x != s]
        object_triples = [(s, p, o)] + [(s, p, x) for x in entity_set if x != o]

        subject_triple_scores = np.array(scoring_function(subject_triples))
        object_triple_scores = np.array(scoring_function(object_triples))

        subject_rank = 1 + np.argsort(np.argsort(- subject_triple_scores))[0]
        object_rank = 1 + np.argsort(np.argsort(- object_triple_scores))[0]

        subject_ranks.append(subject_rank)
        object_ranks.append(object_rank)

        if true_triples is None:
            true_triples = []

        for idx, triple in enumerate(subject_triples):
            if triple != (s, p, o) and triple in true_triples:
                subject_triple_scores[idx] = - np.inf

        for idx, triple in enumerate(object_triples):
            if triple != (s, p, o) and triple in true_triples:
                object_triple_scores[idx] = - np.inf

        subject_rank_filtered = 1 + np.argsort(np.argsort(- subject_triple_scores))[0]
        object_rank_filtered = 1 + np.argsort(np.argsort(- object_triple_scores))[0]

        subject_ranks_filtered.append(subject_rank_filtered)
        object_ranks_filtered.append(object_rank_filtered)

    return (subject_ranks, object_ranks), (subject_ranks_filtered, object_ranks_filtered)


def ranking_summary(res, n=10, tag=None):
    dres = {'subject': dict(), 'object': dict(), 'all': dict()}

    dres['subject']['mean_rank'] = np.mean(res[0])
    dres['subject']['median_rank'] = np.median(res[0])
    dres['subject']['hits@1'] = np.mean(np.asarray(res[0]) <= 1)
    dres['subject']['hits@' + str(n)] = np.mean(np.asarray(res[0]) <= n)
    dres['object']['mean_rank'] = np.mean(res[1])
    dres['object']['median_rank'] = np.median(res[1])
    dres['object']['hits@1'] = np.mean(np.asarray(res[1]) <= 1)
    dres['object']['hits@' + str(n)] = np.mean(np.asarray(res[1]) <= n)

    resg = res[0] + res[1]

    dres['all']['mean_rank'] = np.mean(resg)
    dres['all']['median_rank'] = np.median(resg)
    dres['all']['hits@1'] = np.mean(np.asarray(resg) <= 1)
    dres['all']['hits@' + str(n)] = np.mean(np.asarray(resg) <= n)

    dres['subject']['mrr'] = np.mean(1. / np.asarray(res[0]))
    dres['object']['mrr'] = np.mean(1. / np.asarray(res[1]))
    dres['all']['mrr'] = np.mean(1. / np.asarray(resg))

    return dres



# -*- coding: utf-8 -*-

from setuptools import find_packages
from setuptools import setup
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install


def spacy_download_en():
    import spacy
    try:
        spacy.load('en')
    except:
        import subprocess
        args = ['python3 -m spacy download en']
        subprocess.call(args, shell=True)


class Install(_install):
    def run(self):
        _install.do_egg_install(self)
        spacy_download_en()
        _install.run(self)


class Develop(_develop):
    def run(self):
        spacy_download_en()
        _develop.run(self)


with open('requirements.txt', 'r') as f:
    install_requires = [l for l in f.readlines() if not l.startswith('http://')]

extras_require = {
    'tf': ['tensorflow==1.8.0'],
    'tf_gpu': ['tensorflow-gpu==1.8.0'],
    'torch': ['torch']
}

setup(name='uclmr-jack',
      version='0.2.0',
      description='Jack the Reader is a Python framework for Machine Reading',
      author='UCL Machine Reading',
      author_email='s.riedel@cs.ucl.ac.uk',
      url='https://github.com/uclmr/jack',
      test_suite='tests',
      license='MIT',
      packages=find_packages(),
      cmdclass={
          'install': Install,
          'develop': Develop
      },
      install_requires=install_requires,
      extras_require=extras_require,
      setup_requires=install_requires,
      tests_require=install_requires,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Artificial Intelligence'
      ],
      keywords='tensorflow machine learning natural language processing question answering')



"""

Rajpurkar, Pranav, et al.
"Squad: 100,000+ questions for machine comprehension of text."
arXiv preprint arXiv:1606.05250 (2016).

Original paper: https://arxiv.org/abs/1606.05250
Data: https://rajpurkar.github.io/SQuAD-explorer/
JTR download script: data/SQuAD/download.sh

"""

import argparse
import json


def create_snippet(file_path):
    """
    Creates a snippet of the original SQuAD data.

    Args:
        file_path: path to the original file

    Returns: string containing file contents
    """
    with open(file_path) as data_file:
        data = json.load(data_file)['data']
        out = {
            'data': [{
                'title': data[0]['title'],
                'paragraphs': [{
                    'context': data[0]['paragraphs'][0]['context'],
                    'qas': data[0]['paragraphs'][0]['qas'][0:3]
                }]
            }]
        }
        return json.dumps(out, indent=2)


def create_jtr_snippet(jtr_dict, num_instances=1):
    """
    Creates a jack format snippet from SQuAD data.

    Args:
        jtr_dict: jack dictionary
        num_instances: number of (first) instances

    Returns: dictionary in jack format
    """
    out = dict()
    out['meta'] = jtr_dict['meta']
    out['instances'] = jtr_dict['instances'][0:num_instances]
    return out


def convert_squad(file_path):
    """
    Converts SQuAD dataset to jack format.

    Args:
        file_path: path to the SQuAD json file (train-v1.1.json and dev-v1.1.json in data/SQuAD/)

    Returns: dictionary in jack format
    """
    # meta info
    if '/' in file_path:
        filename = file_path[file_path.rfind('/')+1:]   # Maybe support a system-specific delimiter
    else:
        filename = file_path
    # data
    question_sets = []
    with open(file_path) as data_file:
        data = json.load(data_file)['data']
        for article in data:
            for paragraph in article['paragraphs']:
                qa_set = {
                    'support': [__parse_support(paragraph)],
                    'questions': [__parse_question(qa_dict) for qa_dict in paragraph['qas']]
                }
                question_sets.append(qa_set)
    corpus_dict = {
        'meta': {
            'source': filename
        },
        'instances': question_sets
    }
    return corpus_dict


def __parse_support(para_dict):
    return {
            'text': para_dict['context']
    }


def __parse_question(qa_dict):
    answers = [__parse_answer(answer_dict) for answer_dict in qa_dict['answers']]
    return {
        'question': {
            'text': qa_dict['question'],
            'id': qa_dict['id']
        },
        'answers': answers
    }


def __parse_answer(answer_dict):
    answer_text = answer_dict['text']
    answer_start = answer_dict['answer_start']
    answer_end = answer_start + len(answer_text)
    return {
        'text': answer_text,
        'span': (answer_start, answer_end),
        'doc_idx': 0,  # in SQuAD there is always only a single document
    }


def main():
    """
    Main call function

    Usage:
        from other code:
            call convert_squad(filename)
        from command line:
            call with --help for help

    Returns: nothing
    """
    parser = argparse.ArgumentParser(description='SQuAD dataset to jack format converter.')
    parser.add_argument('infile',
                        help="path to the input file, original SQuAD file (e.g. data/SQuAD/train-v1.1.json)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/SQuAD/train.jack.json)")
    parser.add_argument('-s', '--snippet', action="store_true",
                        help="Export a snippet (first 5 instances) instead of the full file")

    args = parser.parse_args()
    corpus = convert_squad(args.infile)
    if args.snippet:
        corpus = create_jtr_snippet(corpus, num_instances=5)

    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)


if __name__ == "__main__":
    main()



import os


def readAnn(textfolder="../data/SemEval2017Task10/"):
    '''
    Read .ann files and look up corresponding spans in .txt files
    
    Args:
        textfolder:
    '''

    flist = os.listdir(textfolder)
    for f in flist:
        if not f.endswith(".ann"):
            continue

        f_anno = open(os.path.join(textfolder, f), "rU")
        f_text = open(os.path.join(textfolder, f.replace(".ann", ".txt")), "rU")

        # there's only one line, as each .ann file is one text paragraph
        for l in f_text:
            text = l

        #@TODO: collect all keyphrase and relation annotations, create pairs of all keyphrase that appear in same sentence for USchema style RE

        for l in f_anno:
            anno_inst = l.strip().split("\t")
            if len(anno_inst) == 3:
                keytype, start, end = anno_inst[1].split(" ")
                if not keytype.endswith("-of"):

                    # look up span in text and print error message if it doesn't match the .ann span text
                    keyphr_text_lookup = text[int(start):int(end)]
                    keyphr_ann = anno_inst[2]
                    if keyphr_text_lookup != keyphr_ann:
                        print("Spans don't match for anno " + l.strip() + " in file " + f)

                #if keytype.endswith("-of"):


if __name__ == '__main__':
    readAnn()


"""

Hill, Felix, et al.
"The Goldilocks Principle: Reading Children's Books with Explicit Memory Representations."
arXiv preprint arXiv:1511.02301 (2015).


Original paper: https://arxiv.org/abs/1511.02301
Data: https://research.fb.com/projects/babi/
JTR download script: data/CBT/download.sh

"""

import json
import argparse


def __load_cbt_file(path=None, part='train', mode='NE'):
    """
    NOT USED RIGHT NOW

    Path should be given and function will load raw data.
    If it is no given however, there's three parts:
    'train', 'valid' and 'test' as well as 5 modes.
    The modes are:
            - 'CN' (predicting common nouns)
            - 'NE' (predicting named entities)
            - 'P'  (predicting prepositions)
            - 'V'  (predicting verbs.)
            - 'all'(all of the above four categories)
    When calling this function both the dataset part and the mode have to
    be specified.
    'all' is not suitable for QA format, it seems to be just raw text.

    Args:
        path:
        part:
        mode:

    Returns:

    """
    if path is None:
        if mode == 'all':
            path += 'cbt_' + part + '.txt'
        else:
            path += 'cbtest_' + mode + '_' + part
            if part == 'valid':
                path += '_2000ex.txt'
            elif part == 'test':
                path += '_2500ex.txt'
    with open(path, 'r') as f:
        data = f.read()
    return data


def __split_cbt(raw_data, first_n=None):
    """ splits raw cbt data into parts corresponding to each instance """
    story_instances = []
    instance = []
    for l in raw_data.split('\n')[:-1]:
        if l == '':  # reset instance every time an empty line is encountered
            story_instances.append(instance)
            instance = []
            continue
        instance.append(l)
    if first_n:
        return story_instances[:first_n]
    return story_instances


def __parse_cbt_example(instance):
    support = question = answer = candidates_string = []
    for line in instance:
        line_number, line_content = line.split(" ", 1)
        if int(line_number) < 21:    # line contains sentence
            support.append(line_content)
        else:
            question, answer, candidates_string = line_content.split('\t', 2)
    candidates_list = candidates_string.strip('\t').split('|')
    qdict = {
        'question': question,
        'candidates': [{'text': cand} for cand in candidates_list],
        'answers': [{'text': answer}]
    }
    questions = [qdict]
    qset_dict = {
        'suport': [{'text': supp} for supp in support],
        'questions': questions
    }

    return qset_dict


def create_jtr_snippet(path, n_instances=5):
    """
    Creates a jack format snippet.

    Args:
        path: path to the file
        n_instances: number of instances

    Returns: jack json

    """
    return convert_cbt(path, n_instances)


def convert_cbt(path, n_instances=None):
    """
    Convert the Children's Book Test file into jack format.

    Args:
        path: the file which should be converted
        n_instances: how many instances to filter out
    Returns: dictionary in jack format

    """
    # raw_data = __load_cbt_file(path)
    with open(path, 'r') as f:
        raw_data = f.read()

    instances = __split_cbt(raw_data, n_instances)

    corpus = []
    for inst in instances:
        corpus.append(__parse_cbt_example(inst))

    return {
        'meta': 'Children\'s Book Test',
        'globals': {'candidates': []},
        'instances': corpus
    }


def main():
    """
        Main call function

    Usage:
        from other code:
            call convert_cbt(filename)
        from command line:
            call with --help for help

    Returns: nothing
    """

    parser = argparse.ArgumentParser(description='The Children’s Book Test (CBT) dataset to jack format converter.')
    parser.add_argument('infile',
                        help="path to the file to be converted (e.g. data/CBT/CBTest/data/cbtest_CN_train.txt)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/CBT/train.jack.json)")
    parser.add_argument('-s', '--snippet', action="store_true",
                        help="Export a snippet (first 5 instances) instead of the full file")
    args = parser.parse_args()

    if args.snippet:
        corpus = convert_cbt(args.infile, n_instances=5)
    else:
        corpus = convert_cbt(args.infile)
    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)

if __name__ == "__main__":
    main()



"""
This files allows creating a jtr datafile for the SNLI corpus,
whereby
- questions consist of premise + delim. + hypothesis
- TODO: support facts can optinally be loaded from Wordnet / SNLItraining / PPDB
"""


import json

__candidate_labels = ['entailment','neutral','contradiction']
__candidates = [{'text':cl} for cl in __candidate_labels]

CONJ = '|||||'


def convert_snli(snli_file_jsonl, support=False):
    """ io SNLI files into jack format.
    Data source: http://nlp.stanford.edu/projects/snli/snli_1.0.zip
    Files to be converted: snli_1.0_dev.jsonl, snli_1.0_train.jsonl, snli_1.0_test.jsonl
    (the *.txt files contain the same data in a different format)

    Format:
        - question = the premise + CONJ + the hypothesis  = 'sentence1' + CONJ + 'sentence2'
        - support instance = part1 + CONJ + part2 + CONJ + label
    Notes:
        - instances with gold labels '-' are removed from the corpus

    Args:
        snli_file_json: input file
        support: False (no support), TODO: 'WordNet', 'SNLItrain', 'PPDB'
    """
    assert 'snli_1.0' in snli_file_jsonl and snli_file_jsonl.endswith('.jsonl'), "input should be the snli_1.0_X.jsonl files (X=test/train/dev)"

    with open(snli_file_jsonl,'r') as f:
        data = [__convert_snli_instance(json.loads(line.strip())) for line in f.readlines()]
        instances = __add_support([d for d in data if d], support)

        return {'meta': 'SNLI',
                'globals': {'candidates': __candidates},
                'instances': instances  # filter out invalid ones
                }


def __convert_snli_instance(instance):
    try:
        if not instance['gold_label'] in __candidate_labels:
            raise IOError('invalid gold label')
        queb = {}
        queb['id'] = instance['pairID']
        queb['support'] = []
        queb['questions'] = [
            {'question': instance['sentence1'] + CONJ + instance['sentence2'],
             'answers': [
                 {'text': __candidate_labels[__candidate_labels.index(instance['gold_label'])]}]}]

        return queb

    except IOError:
        return None


def __add_support(instances, support):
    """
    Args:
        instances: list of jack instances (with or without support)
        support: False (no support), TODO: 'WordNet', 'SNLItrain', 'PPDB'
    """

    if support in ['WordNet', 'SNLItrain', 'PPDB']:
        # TODO: add support
        pass

    return instances


def main():
    import sys
    if len(sys.argv) == 2:
        corpus = convert_snli(sys.argv[1])
        print(json.dumps(corpus, indent=2))
    else:
        for corpus_name in ["dev","train","test"]:
            corpus = convert_snli("./jack/data/SNLI/snli_1.0/snli_1.0_%s.jsonl" % corpus_name, support=False)
            with open("./jack/data/SNLI/snli_1.0/snli_1.0_%s_jtr_v2.json" % corpus_name, 'w') as outfile:
                json.dump(corpus, outfile, indent=2)

        # create snippet
        corpus = convert_snli("./jack/data/SNLI/snli_1.0/snli_1.0_train.jsonl", support=False)
        corpus['instances'] = corpus['instances'][:10]
        with open("./jack/data/SNLI/snli_1.0/snli_1.0_debug_jtr_v2.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)
        with open("./jack/data/SNLI/snippet_jtrformat_v2.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

if __name__ == "__main__":
    main()



"""

Hermann, Karl Moritz, et al.
"Teaching machines to read and comprehend."
Advances in Neural Information Processing Systems. 2015.

Original paper: https://arxiv.org/abs/1506.03340
Data:   https://github.com/deepmind/rc-data
        http://cs.nyu.edu/~kcho/DMQA/       (direct download)
JTR download script: no download script, check README.md

Metadata:

Number of questions:

        CNN     DailyMail

train   380298  879450
dev     3924    64835
test    3198    53182

"""

import json
import os
import argparse


def create_jtr_snippet(directory, dataset, split, resolve_entities=False, num_instances=5):
    """
    Creates a jack format snippet from rc-data data.

    Args:
        directory: root directory of rc-data
        dataset: which dataset, 'cnn' of 'dailymail'
        split: 'train', 'dev' or 'test'
        resolve_entities: whether to de-anonymise entities. Default: False
        num_instances: number of (first) instances

    Returns:
    """
    return convert_rcdata(directory, dataset, split, resolve_entities, num_instances)


def convert_rcdata(directory, dataset, split, resolve_entities=False, first_n=None):
    """
    Convert subset of rc-data (definet by a combination of dataset and mode) to jtk format
    Args:
        directory: root directory of rc-data
        dataset: which dataset, 'cnn' of 'dailymail'
        split: 'train', 'dev' or 'test'
        resolve_entities: whether to de-anonymise entities. Default: False
        first_n: export a snippet containing the first n instances of the dataset
    Returns:
        jack json

    """
    split_mapping = {'train': 'training', 'test': 'test', 'dev': 'validation'}
    assert split in split_mapping.keys()
    assert dataset in {'cnn', 'dailymail'}

    if directory[-1] != '/':
        directory += "/"
    directory += "{0}/questions/{1}/".format(dataset, split_mapping[split])
    filenames = [file for file in os.listdir(directory) if file.endswith('question')]

    data = {}
    i = 0
    for fname in filenames:
        with open(directory + fname, 'r') as f:
            url = f.readline().strip()
            f.readline()
            text = f.readline().strip()
            f.readline()
            cloze_q = f.readline().strip()
            f.readline()
            answer = f.readline().strip()
            f.readline()
            if resolve_entities:
                entity_dict = {}
                for line in f:
                    split = line.rstrip().split(sep=':')
                    if len(split) != 2:
                        print('Bad split')
                    entity_dict[split[0]] = split[1]
                for key, value in entity_dict.items():
                    text = text.replace(key, value)
                    cloze_q = cloze_q.replace(key, value)
                    answer = answer.replace(key, value)

            if url in data:
                data[url]['rest'].append((fname, cloze_q, answer))
            else:
                data[url] = {
                    'text': text,
                    'rest': [(fname, cloze_q, answer)]
                }
        i += 1
        if first_n and i == first_n:
            break

    instances = []
    counter = 0
    for k, v in data.items():
        url = k
        text = v['text']
        questions = []
        for elem in v['rest']:
            fname, cloze_q, answer = elem
            questions.append({
                'question': cloze_q,
                'answers': [{
                    'text': answer
                }],
                'id': fname
            })
            counter += 1

        instance = {
            # 'id': fname,
            'support': {
                "id": url,
                "text": text
            },
            'questions': questions
        }
        instances.append(instance)

    print(' ...loaded {0} questions.'.format(counter))
    return {
        "meta": "{0}_{1}".format(dataset, split),
        "instances": instances
    }


def main():
    """
    Main call function

    Usage:
        from other code:
            call convert_rcdata(filename)
        from command line:
            call with --help for help

    Returns: nothing
    """
    parser = argparse.ArgumentParser(description='rc-data datasets to jack format converter.')
    parser.add_argument('indir',
                        help="path to the rc-data root directory (e.g. data/rc-data/)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/rc-data/cnn_train.jack.json)")
    parser.add_argument('dataset', choices=['cnn', 'dailymail'],
                        help="which dataset to access: cnn or dailymail")
    parser.add_argument('split', choices=['train', 'dev', 'test'],
                        help="which split of the dataset to io: train, dev or test")
    parser.add_argument('-s', '--snippet', action="store_true",
                        help="Export a snippet (first 5 instances) instead of the full file")
    args = parser.parse_args()

    if args.snippet:
        corpus = create_jtr_snippet(args.indir, args.dataset, args.split, num_instances=5)
    else:
        corpus = convert_rcdata(args.indir, args.dataset, args.split)

    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)


if __name__ == "__main__":
    main()



import tensorflow as tf

from jack.core import TensorPortTensors, TensorPort
from jack.readers.extractive_qa.tensorflow.abstract_model import AbstractXQAModelModule
from jack.readers.extractive_qa.tensorflow.answer_layer import answer_layer
from jack.util.tf.embedding import conv_char_embedding
from jack.util.tf.modular_encoder import modular_encoder


class ModularQAModel(AbstractXQAModelModule):
    def set_topk(self, k):
        self._topk_assign(k)

    def create_output(self, shared_resources, input_tensors):
        tensors = TensorPortTensors(input_tensors)

        [char_emb_question, char_emb_support] = conv_char_embedding(
            len(shared_resources.char_vocab), shared_resources.config['repr_dim'], tensors.word_chars,
            tensors.word_char_length, [tensors.question_batch_words, tensors.support_batch_words])

        model = shared_resources.config['model']
        repr_dim = shared_resources.config['repr_dim']
        input_size = shared_resources.embeddings.shape[-1]
        dropout = shared_resources.config.get("dropout")
        tensors.emb_question.set_shape([None, None, input_size])
        tensors.emb_support.set_shape([None, None, input_size])

        inputs = {'question': tensors.emb_question, 'support': tensors.emb_support,
                  'char_question': char_emb_question, 'char_support': char_emb_support,
                  'word_in_question': tf.expand_dims(tensors.word_in_question, 2)}
        inputs_length = {'question': tensors.question_length, 'support': tensors.support_length,
                         'char_question': tensors.question_length, 'char_support': tensors.support_length,
                         'word_in_question': tensors.support_length}
        inputs_mapping = {'question': None, 'support': tensors.support2question,
                          'char_support': tensors.support2question}

        encoder_config = model['encoder_layer']

        encoded, lengths, mappings = modular_encoder(
            encoder_config, inputs, inputs_length, inputs_mapping, repr_dim, dropout, tensors.is_eval)

        with tf.variable_scope('answer_layer'):
            answer_layer_config = model['answer_layer']
            encoded_question = encoded[answer_layer_config.get('question', 'question')]
            encoded_support = encoded[answer_layer_config.get('support', 'support')]

            if 'repr_dim' not in answer_layer_config:
                answer_layer_config['repr_dim'] = repr_dim
            if 'max_span_size' not in answer_layer_config:
                answer_layer_config['max_span_size'] = shared_resources.config.get('max_span_size', 16)
            topk = tf.get_variable(
                'topk', initializer=shared_resources.config.get('topk', 1), dtype=tf.int32, trainable=False)
            topk_p = tf.placeholder(tf.int32, [], 'beam_size_setter')
            topk_assign = topk.assign(topk_p)
            self._topk_assign = lambda k: self.tf_session.run(topk_assign, {topk_p: k})

            start_scores, end_scores, doc_idx, predicted_start_pointer, predicted_end_pointer = \
                answer_layer(encoded_question, lengths[answer_layer_config.get('question', 'question')],
                             encoded_support, lengths[answer_layer_config.get('support', 'support')],
                             mappings[answer_layer_config.get('support', 'support')],
                             tensors.answer2support, tensors.is_eval,
                             tensors.correct_start, topk=topk, **answer_layer_config)

        span = tf.stack([doc_idx, predicted_start_pointer, predicted_end_pointer], 1)

        return TensorPort.to_mapping(self.output_ports, (start_scores, end_scores, span))



import json
from collections import defaultdict
import sys
import os


def main():
    # = parse_cbt_example(instances[0])
    if len(sys.argv) == 2:
        with open(sys.argv[1], 'r') as f:
            sentihood_data = json.load(f)

        convert_to_jtr(sentihood_data)
    elif len(sys.argv) ==1:
        data_path = '../data/sentihood/'
        filenames = ['sentihood-train.json', 'sentihood-dev.json',
        'sentihood-test.json']
        for i, f in enumerate(filenames):
            raw_data = json.load(open(os.path.join(data_path, f)))
            instances = convert_to_jtr(raw_data)

            if i == 0: # training data -> write overfit set
                json.dump(wrap_into_jtr_global(instances[:100]),
                        open('../../tests/test_data/sentihood/overfit.json','w'),
                        indent=2)

            # write data sets for smalldata tests
            json.dump(wrap_into_jtr_global(instances[:1000]),
                    open(os.path.join('../../tests/test_data/sentihood/',f),'w'),
                    indent=2)

def wrap_into_jtr_global(instances):
    reading_dataset = {
        'globals': {
            'candidates': [
                {'text': 'Negative'},
                {'text': 'Positive'},
                {'text': 'Neutral'}
            ]
        },
        'instances': instances
    }
    return reading_dataset



def convert_to_jtr(sentihood_data, exhaustive=True):
    instances = []
    # collect all aspects
    aspects = set()
    for instance in sentihood_data:
        if 'opinions' in instance.keys():
            for opinion in instance['opinions']:
                aspects.add(opinion['aspect'])
    for instance in sentihood_data:
        text = instance['text']
        answers = defaultdict(lambda: 'Neutral')
        if 'opinions' in instance.keys():
            for opinion in instance['opinions']:
                aspect = opinion['aspect']
                answers[aspect] = opinion['sentiment']

        for aspect in aspects if exhaustive else answers.keys():
            reading_instance = {
                'support': [{'text': text}],
                'questions': [{'question': aspect, 'answers': [{'text': answers[aspect]}]}]
            }
            instances.append(reading_instance)

    return instances


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from abc import *
from collections import defaultdict
from datetime import datetime
from time import strftime, localtime
from time import time
from typing import List, Tuple, Mapping

import numpy as np
import progressbar
import tensorflow as tf

from jack.core.data_structures import QASetting, Answer
from jack.core.reader import JTReader
from jack.core.tensorflow import TFReader
from jack.core.tensorport import TensorPort, Ports
from jack.eval.extractive_qa import *

logger = logging.getLogger(__name__)

"""
TODO -- hooks should also have prefixes so that one can use the same hook with different parameters
"""


class TrainingHook(metaclass=ABCMeta):
    """Serves as Hook interface."""

    @abstractmethod
    def reader(self) -> JTReader:
        """ Returns: JTReader instance"""
        raise NotImplementedError

    @abstractmethod
    def at_epoch_end(self, epoch: int, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def at_iteration_end(self, epoch: int, loss: float, set_name='train', **kwargs):
        raise NotImplementedError


class TFTrainingHook(TrainingHook):
    """Serves as Hook interface."""

    @abstractmethod
    def reader(self) -> TFReader:
        """ Returns: JTReader instance"""
        raise NotImplementedError


class TraceHook(TFTrainingHook):
    """Abstract hook class, which implements an update function the summary."""

    def __init__(self, reader, summary_writer=None):
        self._summary_writer = summary_writer
        self._reader = reader
        self.scores = {}

    @property
    def reader(self) -> TFReader:
        return self._reader

    def update_summary(self, current_step, title, value):
        """Adds summary (title, value) to summary writer object.

        Args:
            current_step (int): Current step in the training procedure.
            value (float): Scalar value for the message.
        """
        if self._summary_writer is not None:
            summary = tf.Summary(value=[
                tf.Summary.Value(tag=title, simple_value=value),
            ])
            self._summary_writer.add_summary(summary, current_step)

    def add_to_history(self, score_dict, iter_value, epoch, set_name='train'):
        for metric in score_dict:
            if metric not in self.scores: self.scores[metric] = {}
            if set_name not in self.scores[metric]: self.scores[metric][set_name] = [[], [], []]
            self.scores[metric][set_name][0].append(score_dict[metric])
            self.scores[metric][set_name][1].append(iter_value)
            self.scores[metric][set_name][2].append(epoch)


class LossHook(TraceHook):
    """A hook at prints the current loss and adds it to the summary."""

    def __init__(self, reader, iter_interval=None, summary_writer=None):
        super(LossHook, self).__init__(reader, summary_writer)
        self._iter_interval = iter_interval
        self._acc_loss = {'train': 0.0}
        self._iter = {'train': 0}
        self._epoch_loss = {'train': 0.0}
        self._iter_epoch = {'train': 0}

    def at_iteration_end(self, epoch, loss, set_name='train', **kwargs):
        """Prints the loss, epoch, and #calls; adds it to the summary. Loss should be batch normalized."""
        if self._iter_interval is None: return loss
        if set_name not in self._acc_loss:
            self._acc_loss[set_name] = 0.0
            self._iter[set_name] = 0
            self._epoch_loss[set_name] = 0.0
            self._iter_epoch[set_name] = 0

        self._iter_epoch[set_name] += 1
        self._epoch_loss[set_name] += 1
        self._iter[set_name] += 1
        self._acc_loss[set_name] += loss

        if not self._iter[set_name] == 0 and self._iter[set_name] % self._iter_interval == 0:
            loss = self._acc_loss[set_name] / self._iter_interval
            super().add_to_history({'loss': loss, },
                                   self._iter[set_name], epoch, set_name)
            logger.info("Epoch {0}\tIter {1}\t{3} loss {2}".format(epoch,
                                                                   self._iter[set_name], loss, set_name))
            self.update_summary(self._iter[set_name], "{0} loss".format(set_name), loss)
            self._acc_loss[set_name] = 0

        ret = (0.0 if self._iter[set_name] == 0 else self._acc_loss[set_name] / self._iter[set_name])

        return ret

    def at_epoch_end(self, epoch, set_name='train', **kwargs):
        if self._iter_interval is None:
            loss = self._acc_loss[set_name] / self._iter_interval[set_name]
            logger.info("Epoch {}\tIter {}\t{3} Loss {}".format(epoch,
                                                                self._iter, loss, set_name))
            self.update_summary(self._iter[set_name], "Loss", loss)
            self._epoch_loss[set_name] = 0
            self._iter_epoch[set_name] = 0

        ret = (0.0 if self._iter_epoch[set_name] == 0 else self._epoch_loss[set_name] / self._iter_epoch[set_name])

        return ret


class ExamplesPerSecHook(TraceHook):
    """Prints the examples per sec and adds it to the summary writer."""

    def __init__(self, reader, batch_size, iter_interval=None, summary_writer=None):
        super(ExamplesPerSecHook, self).__init__(reader, summary_writer)
        self._iter_interval = iter_interval
        self._iter = 0
        self.num_examples = iter_interval * batch_size
        self.reset = True

    def __tag__(self):
        return "Speed"

    def at_epoch_end(self, epoch, **kwargs):
        # to eliminate drop in measured speed due to post-epoch hooks:
        # do not execute; reset for use during epochs only
        self.reset = True

    def at_iteration_end(self, epoch, loss, **kwargs):
        """Prints the examples per sec and adds it to the summary writer."""
        self._iter += 1
        if self.reset:
            self.t0 = time()
            self.reset = False
        elif self._iter % self._iter_interval == 0:
            diff = time() - self.t0
            speed = "%.2f" % (self.num_examples / diff)
            logger.info("Epoch {}\tIter {}\tExamples/s {}".format(str(epoch), str(self._iter), str(speed)))
            self.update_summary(self._iter, self.__tag__(), float(speed))
            self.t0 = time()


class ETAHook(TraceHook):
    """Estimates ETA to next checkpoint, epoch end and training end."""

    def __init__(self, reader, iter_interval, iter_per_epoch, max_epochs, iter_per_checkpoint=None,
                 summary_writer=None):
        super(ETAHook, self).__init__(reader, summary_writer)
        self.iter_interval = iter_interval
        self.iter_per_epoch = iter_per_epoch
        self.iter_per_checkpoint = iter_per_checkpoint
        self.iter = 0
        self.epoch = 1
        self.max_epochs = max_epochs
        self.max_iters = max_epochs * iter_per_epoch
        self.start = time()
        self.start_checkpoint = time()
        self.start_epoch = time()
        self.reestimate = True

    def __tag__(self):
        return "ETA"

    def at_epoch_end(self, epoch, **kwargs):
        # to eliminate drop in measured speed due to post-epoch hooks:
        # do not execute; reset for use during epochs only
        self.start_epoch = time()

    def at_iteration_end(self, epoch, loss, **kwargs):
        """Estimates ETA from max_iter vs current_iter."""
        self.iter += 1

        def format_eta(seconds):
            if seconds == float("inf"):
                return "never"
            else:
                seconds, _ = divmod(seconds, 1)
                minutes, seconds = divmod(seconds, 60)
                hours, minutes = divmod(minutes, 60)
                seconds = str(int(seconds))
                minutes = str(int(minutes))
                hours = str(int(hours))

                if len(hours) < 2:
                    hours = "0" + hours
                if len(minutes) < 2:
                    minutes = "0" + minutes
                if len(seconds) < 2:
                    seconds = "0" + seconds

                return "{}:{}:{}".format(hours, minutes, seconds)

        if not self.iter == 0 and self.iter % self.iter_interval == 0:
            current_time = time()

            def get_eta(progress, start_time, name):
                elapsed = current_time - start_time
                eta = elapsed / progress * (1.0 - progress)
                eta_date = strftime("%y-%m-%d %H:%M:%S", localtime(current_time + eta))
                self.update_summary(self.iter, self.__tag__() + "_" + name, float(eta))

                return format_eta(eta), eta_date

            log = "Epoch %d\tIter %d" % (epoch, self.iter)
            total_progress = float(self.iter) / self.max_iters
            eta, eta_data = get_eta(total_progress, self.start, "total")
            log += "\tETA: %s, %s (%.2f%%)" % (eta, eta_data, total_progress * 100)
            epoch_progress = float((self.iter - 1) % self.iter_per_epoch + 1) / self.iter_per_epoch
            eta, _ = get_eta(epoch_progress, self.start_epoch, "epoch")
            log += "\tETA(epoch): %s (%.2f%%)" % (eta, epoch_progress * 100)
            if self.iter_per_checkpoint is not None:
                checkpoint_progress = float((self.iter - 1) % self.iter_per_checkpoint + 1) / self.iter_per_checkpoint
                eta, _ = get_eta(checkpoint_progress, self.start_checkpoint, "checkpoint")
                log += "\tETA(checkpoint): %s (%.2f%%)" % (eta, checkpoint_progress * 100)

            logger.info(log)

        if self.iter_per_checkpoint is not None and self.iter % self.iter_per_checkpoint == 0:
            self.start_checkpoint = time()


class EvalHook(TraceHook):
    def __init__(self, reader: JTReader, dataset, batch_size: int, ports: List[TensorPort],
                 iter_interval=None, epoch_interval=1, metrics=None, summary_writer=None,
                 write_metrics_to=None, info="", side_effect=None):
        super(EvalHook, self).__init__(reader, summary_writer)
        self._total = len(dataset)
        self._dataset = dataset
        self._batches = None
        self._ports = ports
        self._epoch_interval = epoch_interval
        self._iter_interval = iter_interval
        self._batch_size = batch_size
        # self.done_for_epoch = False
        self._iter = 0
        self._info = info or self.__class__.__name__
        self._write_metrics_to = write_metrics_to
        self._metrics = metrics or self.possible_metrics
        self._side_effect = side_effect
        self._side_effect_state = None

    @abstractmethod
    def possible_metrics(self) -> List[str]:
        """Returns: list of metric keys this evaluation hook produces. """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def preferred_metric_and_initial_score():
        """Returns: Tuple of preferred metric to optimize and its initial value (usually the lowest possible one)"""
        raise NotImplementedError

    @abstractmethod
    def apply_metrics(self, inputs: List[Tuple[QASetting, List[Answer]]], tensors: Mapping[TensorPort, np.ndarray]) \
            -> Mapping[str, float]:
        """Returns: dict from metric name to float"""
        raise NotImplementedError

    def combine_metrics(self, accumulated_metrics: Mapping[str, List[float]]) -> Mapping[str, float]:
        """Returns:
               dict from metric name to float. Per default batch metrics are simply averaged by
               total number of examples"""
        return {k: sum(vs) / self._total for k, vs in accumulated_metrics.items()}

    def __call__(self, epoch):
        if self._batches is None:
            logger.info("Preparing evaluation data...")
            self._batches = self.reader.input_module.batch_generator(self._dataset, self._batch_size, is_eval=True)

        logger.info("Started evaluation %s" % self._info)
        metrics = defaultdict(lambda: list())
        bar = progressbar.ProgressBar(
            max_value=len(self._dataset) // self._batch_size + 1,
            widgets=[' [', progressbar.Timer(), '] ', progressbar.Bar(), ' (', progressbar.ETA(), ') '])
        for i, batch in bar(enumerate(self._batches)):
            inputs = self._dataset[i * self._batch_size:(i + 1) * self._batch_size]
            predictions = self.reader.model_module(batch, self._ports)
            m = self.apply_metrics(inputs, predictions)
            for k in self._metrics:
                metrics[k].append(m[k])

        metrics = self.combine_metrics(metrics)
        super().add_to_history(metrics, self._iter, epoch)

        printmetrics = sorted(metrics.keys())
        res = "Epoch %d\tIter %d\ttotal %d" % (epoch, self._iter, self._total)
        for m in printmetrics:
            res += '\t%s: %.3f' % (m, metrics[m])
            self.update_summary(self._iter, self._info + '_' + m, metrics[m])
            if self._write_metrics_to is not None:
                with open(self._write_metrics_to, 'a') as f:
                    f.write("{0} {1} {2:.5}\n".format(datetime.now(), self._info + '_' + m,
                                                      np.round(metrics[m], 5)))
        res += '\t' + self._info
        logger.info(res)

        if self._side_effect is not None:
            self._side_effect_state = self._side_effect(metrics, self._side_effect_state)

    def at_epoch_end(self, epoch: int, **kwargs):
        if self._epoch_interval is not None and epoch % self._epoch_interval == 0:
            self.__call__(epoch)

    def at_iteration_end(self, epoch: int, loss: float, **kwargs):
        self._iter += 1
        if self._iter_interval is not None and self._iter % self._iter_interval == 0:
            self.__call__(epoch)


class XQAEvalHook(EvalHook):
    """This evaluation hook computes the following metrics: exact and per-answer f1 on token basis."""

    def __init__(self, reader: JTReader, dataset: List[Tuple[QASetting, List[Answer]]], batch_size: int,
                 iter_interval=None, epoch_interval=1, metrics=None, summary_writer=None,
                 write_metrics_to=None, info="", side_effect=None,
                 predicted_answer_span_port=Ports.Prediction.answer_span,
                 target_answer_span_port=Ports.Target.answer_span,
                 answer2support_port=Ports.Input.answer2support,
                 support2question_port=Ports.Input.support2question, **kwargs):
        self._predicted_answer_span_port = predicted_answer_span_port
        self._target_answer_span_port = target_answer_span_port
        self._answer2support_port = answer2support_port
        self._support2question_port = support2question_port
        ports = reader.output_module.input_ports
        super().__init__(reader, dataset, batch_size, ports, iter_interval, epoch_interval, metrics, summary_writer,
                         write_metrics_to, info, side_effect)

    @property
    def possible_metrics(self) -> List[str]:
        return ["exact", "f1"]

    @staticmethod
    def preferred_metric_and_initial_score():
        return 'f1', [0.0]

    def apply_metrics(self, inputs: List[Tuple[QASetting, List[Answer]]], tensors: Mapping[TensorPort, np.ndarray]) \
            -> Mapping[str, float]:
        qs = [q for q, a in inputs]
        p_answers = self.reader.output_module(qs, {p: tensors[p] for p in self.reader.output_module.input_ports})

        f1 = exact_match = 0
        for pa, (q, ass) in zip(p_answers, inputs):
            ground_truth = [a.text for a in ass]
            f1 += metric_max_over_ground_truths(f1_score, pa[0].text, ground_truth)
            exact_match += metric_max_over_ground_truths(exact_match_score, pa[0].text, ground_truth)

        return {"f1": f1, "exact": exact_match}


class ClassificationEvalHook(EvalHook):
    def __init__(self, reader: JTReader, dataset: List[Tuple[QASetting, List[Answer]]], batch_size: int,
                 iter_interval=None, epoch_interval=1, metrics=None, summary_writer=None,
                 write_metrics_to=None, info="", side_effect=None,
                 predicted_index_port=Ports.Prediction.candidate_index,
                 target_index_port=Ports.Target.target_index, **kwargs):
        self._predicted_index_port = predicted_index_port
        self._target_index_port = target_index_port
        ports = [self._predicted_index_port, self._target_index_port]

        super().__init__(reader, dataset, batch_size, ports, iter_interval, epoch_interval, metrics, summary_writer,
                         write_metrics_to, info, side_effect)

    @property
    def possible_metrics(self) -> List[str]:
        return ["Accuracy"]

    @staticmethod
    def preferred_metric_and_initial_score():
        return 'Accuracy', [0.0]

    def apply_metrics(self, inputs: List[Tuple[QASetting, List[Answer]]], tensors: Mapping[TensorPort, np.ndarray]) \
            -> Mapping[str, float]:
        labels = tensors[self._target_index_port]
        predictions = tensors[self._predicted_index_port]

        labels_np = np.array(labels)
        acc_exact = np.sum(np.equal(labels_np, predictions))

        return {"Accuracy": acc_exact}


class LogProbEvalHook(EvalHook):
    """Assumes that the loss represents the negative log  probability of the model."""

    def __init__(self, reader: JTReader, dataset: List[Tuple[QASetting, List[Answer]]], batch_size: int,
                 iter_interval=None, epoch_interval=1, metrics=None, summary_writer=None,
                 write_metrics_to=None, info="", side_effect=None, **kwargs):
        ports = [Ports.loss]
        self.epoch = 0
        super().__init__(reader, dataset, batch_size, ports, iter_interval, epoch_interval, metrics, summary_writer,
                         write_metrics_to, info, side_effect)

    @property
    def possible_metrics(self) -> List[str]:
        return ["log_p"]

    @staticmethod
    def preferred_metric_and_initial_score():
        return 'log_p', [float('-inf')]

    def apply_metrics(self, inputs: List[Tuple[QASetting, List[Answer]]], tensors: Mapping[TensorPort, np.ndarray]) \
            -> Mapping[str, float]:
        loss = tensors[Ports.loss]
        return {"log_p": -loss * len(inputs)}

    def at_epoch_end(self, epoch: int, **kwargs):
        self.epoch += 1
        if self._epoch_interval is not None and epoch % self._epoch_interval == 0:
            self.__call__(epoch)



import os
from typing import Union

from jack.core.shared_resources import SharedResources
from jack.core.tensorflow import TFReader
from jack.util.hooks import XQAEvalHook, ClassificationEvalHook

readers = {}
eval_hooks = {}

extractive_qa_readers = {}
classification_readers = {}
nli_readers = {}
link_prediction_readers = {}


def __reader(f):
    readers.setdefault(f.__name__, f)
    return f


def extractive_qa_reader(f):
    __reader(f)
    extractive_qa_readers.setdefault(f.__name__, f)
    eval_hooks.setdefault(f.__name__, XQAEvalHook)
    return f


def classification_reader(f):
    __reader(f)
    classification_readers.setdefault(f.__name__, f)
    eval_hooks.setdefault(f.__name__, ClassificationEvalHook)
    return f


def nli_reader(f):
    nli_readers.setdefault(f.__name__, f)
    return classification_reader(f)


def link_prediction_reader(f):
    from jack.util.hooks import LogProbEvalHook
    __reader(f)
    link_prediction_readers.setdefault(f.__name__, f)
    eval_hooks.setdefault(f.__name__, LogProbEvalHook)
    return f


def reader_from_file(load_dir: str, **kwargs):
    """Loads a reader from a file.

    Args:
        load_dir: directory containing the reader being loaded.
        kwargs: can be used to overwrite the reader configuration if necessary.

    Returns:
        the reader.
    """
    shared_resources = create_shared_resources()
    shared_resources.load(os.path.join(load_dir, "shared_resources"))
    if kwargs:
        shared_resources.config.update(kwargs)
    reader = readers[shared_resources.config["reader"]](shared_resources)
    reader.load_and_setup_modules(load_dir)
    return reader


def create_shared_resources(resources_or_config: Union[dict, SharedResources] = None) -> SharedResources:
    """
    Produces a SharedResources object based on the input.
    Args:
        resources_or_config: either nothing, a configuration dictionary, or shared resources

    Returns: a SharedResources object.
    """
    if resources_or_config is None:
        return SharedResources()
    elif isinstance(resources_or_config, SharedResources):
        return resources_or_config
    else:
        return SharedResources(config=resources_or_config)


# Question Answering


def _tf_extractive_qa_reader(model_module_constructor, resources_or_conf: Union[dict, SharedResources]):
    from jack.readers.extractive_qa.shared import XQAInputModule, XQAOutputModule
    shared_resources = create_shared_resources(resources_or_conf)
    input_module = XQAInputModule(shared_resources)
    model_module = model_module_constructor(shared_resources)
    output_module = XQAOutputModule()
    return TFReader(shared_resources, input_module, model_module, output_module)


@extractive_qa_reader
def fastqa_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a FastQA reader instance (extractive qa model)."""
    from jack.readers.extractive_qa.tensorflow.fastqa import FastQAModule
    return _tf_extractive_qa_reader(FastQAModule, resources_or_conf)


@extractive_qa_reader
def modular_qa_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a FastQA reader instance (extractive qa model)."""
    from jack.readers.extractive_qa.tensorflow.modular_qa_model import ModularQAModel
    return _tf_extractive_qa_reader(ModularQAModel, resources_or_conf)


@extractive_qa_reader
def fastqa_reader_torch(resources_or_conf: Union[dict, SharedResources] = None):
    """ Creates a FastQA reader instance (extractive qa model). """
    from jack.readers.extractive_qa.torch.fastqa import FastQAPyTorchModelModule
    from jack.readers.extractive_qa.shared import XQAInputModule, XQAOutputModule
    from jack.core.torch import PyTorchReader
    shared_resources = create_shared_resources(resources_or_conf)
    input_module = XQAInputModule(shared_resources)
    model_module = FastQAPyTorchModelModule(shared_resources)
    output_module = XQAOutputModule()
    return PyTorchReader(shared_resources, input_module, model_module, output_module)


# Natural Language Inference


def _tf_nli_reader(model_module_constructor, resources_or_conf: Union[dict, SharedResources] = None):
    from jack.readers.classification.shared import ClassificationSingleSupportInputModule
    from jack.readers.classification.shared import SimpleClassificationOutputModule
    shared_resources = create_shared_resources(resources_or_conf)
    input_module = ClassificationSingleSupportInputModule(shared_resources)
    model_module = model_module_constructor(shared_resources)
    output_module = SimpleClassificationOutputModule(shared_resources)
    return TFReader(shared_resources, input_module, model_module, output_module)


@nli_reader
def dam_snli_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a NLI reader instance.
    This particular reader uses a Decomposable Attention Model, as described in [1].
    [1] Ankur P. Parikh et al. - A Decomposable Attention Model for Natural Language Inference. EMNLP 2016
    """
    from jack.readers.natural_language_inference.decomposable_attention import DecomposableAttentionModel
    return _tf_nli_reader(DecomposableAttentionModel, resources_or_conf)


@nli_reader
def cbilstm_nli_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a NLI reader based on conditional BiLSTMs."""
    from jack.readers.natural_language_inference.conditional_bilstm import ConditionalBiLSTMClassificationModel
    return _tf_nli_reader(ConditionalBiLSTMClassificationModel, resources_or_conf)


@nli_reader
def modular_nli_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a Modular NLI reader instance. Model defined in config."""
    from jack.readers.natural_language_inference.modular_nli_model import ModularNLIModel
    return _tf_nli_reader(ModularNLIModel, resources_or_conf)


# Link Prediction/Knowledge Base Population Models


@link_prediction_reader
def distmult_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """Creates a knowledge_base_population DistMult model."""
    from jack.readers.link_prediction.models import KnowledgeGraphEmbeddingInputModule, \
        KnowledgeGraphEmbeddingModelModule, \
        KnowledgeGraphEmbeddingOutputModule
    shared_resources = create_shared_resources(resources_or_conf)
    input_module = KnowledgeGraphEmbeddingInputModule(shared_resources)
    model_module = KnowledgeGraphEmbeddingModelModule(shared_resources, model_name='DistMult')
    output_module = KnowledgeGraphEmbeddingOutputModule()
    return TFReader(shared_resources, input_module, model_module, output_module)


@link_prediction_reader
def complex_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """ Creates a knowledge_base_population Complex model."""
    from jack.readers.link_prediction.models import KnowledgeGraphEmbeddingInputModule, \
        KnowledgeGraphEmbeddingModelModule, \
        KnowledgeGraphEmbeddingOutputModule
    shared_resources = create_shared_resources(resources_or_conf)
    input_module = KnowledgeGraphEmbeddingInputModule(shared_resources)
    model_module = KnowledgeGraphEmbeddingModelModule(shared_resources, model_name='ComplEx')
    output_module = KnowledgeGraphEmbeddingOutputModule()
    return TFReader(shared_resources, input_module, model_module, output_module)


@link_prediction_reader
def transe_reader(resources_or_conf: Union[dict, SharedResources] = None):
    """ Creates a knowledge_base_population TransE model."""
    from jack.readers.link_prediction.models import KnowledgeGraphEmbeddingInputModule, \
        KnowledgeGraphEmbeddingModelModule, \
        KnowledgeGraphEmbeddingOutputModule
    shared_resources = create_shared_resources(resources_or_conf)

    input_module = KnowledgeGraphEmbeddingInputModule(shared_resources)
    model_module = KnowledgeGraphEmbeddingModelModule(shared_resources, model_name='TransE')
    output_module = KnowledgeGraphEmbeddingOutputModule()
    return TFReader(shared_resources, input_module, model_module, output_module)



import argparse
import csv
import json
from collections import Counter


def convert_newsqa(file_path):
    """
    Converts NewsQA dataset to jack format.

    Args:
        file_path: path to the NewsQA CSV file (data/NewsQA/)

    Returns: dictionary in jack format
    """
    # meta info
    if '/' in file_path:
        filename = file_path[file_path.rfind('/') + 1:]  # Maybe support a system-specific delimiter
    else:
        filename = file_path

    # data
    question_sets = []
    with open(file_path) as data_file:
        reader = csv.reader(data_file)
        reader.__next__()
        for row in reader:
            [story_id, question, answer_char_ranges, is_answer_absent, is_question_bad, validated_answers,
             story_text] = row

            if validated_answers:
                answers = json.loads(validated_answers)
                spans = [k for k, v in answers.items() if ":" in k]
            else:
                answers = Counter()
                for rs in answer_char_ranges.split("|"):
                    for r in set(rs.split(",")):
                        if ":" in r:
                            answers[r] += 1
                spans = [k for k, v in answers.items() if ":" in k and v >= 2]

            if spans:
                qa_set = {
                    "support": [story_text],
                    "questions": [{
                        'question': {
                            'text': question,
                            'id': story_id + "_" + question.replace(" ", "_")
                        },
                        'answers': [{"span": [int(span.split(":")[0]), int(span.split(":")[1])],
                                     "text": story_text[int(span.split(":")[0]):int(span.split(":")[1])]
                                     } for span in spans]
                    }]
                }
                question_sets.append(qa_set)

    corpus_dict = {
        'meta': {
            'source': filename
        },
        'instances': question_sets
    }

    return corpus_dict


def main():
    """
    Main call function

    Usage:
        from other code:
            call convert_squad(filename)
        from command line:
            call with --help for help

    Returns: nothing
    """
    parser = argparse.ArgumentParser(description='NewsQA dataset to jack format converter.')
    parser.add_argument('infile',
                        help="path to the input file, original NewsQA file (e.g. data/NewsQA/newsqa/maluuba/newsqa/split_data/train.csv)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/NewsQA/train.jack.json)")

    args = parser.parse_args()
    corpus = convert_newsqa(args.infile)

    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)


if __name__ == "__main__":
    main()



import csv
import json
import sys
from collections import Counter

input_fn = sys.argv[1]
output_fn = sys.argv[2]

dataset = []
squad_style_dataset = {"data": dataset, "version": "1"}

with open(input_fn, "r") as f:
    reader = csv.reader(f)
    reader.__next__()
    for row in reader:
        [story_id, question, answer_char_ranges, is_answer_absent, is_question_bad, validated_answers, story_text] = row

        spans = None
        if validated_answers:
            answers = json.loads(validated_answers)
            spans = [k for k, v in answers.items() if ":" in k]
        else:
            answers = Counter()
            for rs in answer_char_ranges.split("|"):
                for r in set(rs.split(",")):
                    if ":" in r:
                        answers[r] += 1
            spans = [k for k, v in answers.items() if ":" in k and v >= 2]

        if spans:
            example = {"title": story_id, "paragraphs": [
                {
                    "context": story_text,
                    "qas": [{
                        "question": question,
                        "id": story_id + "_" + question.replace(" ", "_"),
                        "answers": [{
                                        "answer_start": int(span.split(":")[0]),
                                        "text": story_text[int(span.split(":")[0]):int(span.split(":")[1])]
                                    } for span in spans]
                    }]
                }
            ]}
            dataset.append(example)
            # else:
            #    print("No span found for %s" % story_id)

with open(output_fn, "w") as f:
    json.dump(squad_style_dataset, f)



# -*- coding: utf-8 -*-

import logging
import os
import random
import tempfile
from abc import abstractmethod
from typing import Iterable, Tuple, List, Mapping, TypeVar, Generic, Optional

import diskcache as dc
import numpy as np

from jack.core.data_structures import QASetting, Answer
from jack.core.shared_resources import SharedResources
from jack.core.tensorport import TensorPort
from jack.util.batch import shuffle_and_batch, GeneratorWithRestart

logger = logging.getLogger(__name__)


class InputModule:
    """An input module processes inputs and turns them into tensors to be processed by the model module.

    Note that all setting up should be done in the setup method, NOT in the constructor. Only use the constructor to
    hand over external variables/states, like `SharedResources`.
    """

    def setup(self):
        """Optionally, sets up the module (if needs setup after loading shared resources for instance).

        Assumes shared resources are fully setup. Usually called after loading and after `setup_from_data` as well."""
        pass

    def setup_from_data(self, data: Iterable[Tuple[QASetting, List[Answer]]]):
        """Optionally, sets up the module based on input data.

        This usually involves setting up vocabularies and other resources. This
        should and is only called before training, not before loading a saved model.

        Args:
            data: a set of pairs of input and answer.
        """
        pass

    @property
    @abstractmethod
    def output_ports(self) -> List[TensorPort]:
        """
        Defines what types of tensors the output module produces in each batch.
        Returns: a list of tensor ports that correspond to the tensor ports in the mapping
        produced by `__call__`. The `batch_generator` method will return bindings for these
        ports and the ones in `training_ports`.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def training_ports(self) -> List[TensorPort]:
        """
        Defines what types of tensor are provided in addition to `output_ports` during training
        in the `batch_generator` function. Typically these will be ports that describe
        the target solution at training time.
        """
        raise NotImplementedError

    @abstractmethod
    def __call__(self, qa_settings: List[QASetting]) -> Mapping[TensorPort, np.ndarray]:
        """
        Converts a list of inputs into a single batch of tensors, consistent with the `output_ports` of this
        module.
        Args:
            qa_settings: a list of instances (question, support, optional candidates)

        Returns:
            A mapping from ports to tensors.

        """
        raise NotImplementedError

    @abstractmethod
    def batch_generator(self, dataset: Iterable[Tuple[QASetting, List[Answer]]], batch_size: int,
                        is_eval: bool) -> Iterable[Mapping[TensorPort, np.ndarray]]:
        """
        Given a training set of input-answer pairs, this method produces an iterable/generator
        that when iterated over returns a sequence of batches. These batches map ports to tensors
        just as `__call__` does, but provides additional bindings for the `training_ports` ports in
        case `is_eval` is `False`.

        Args:
            dataset: a set of pairs of input and answer.
            is_eval: is this dataset generated for evaluation only (not training).

        Returns: An iterable/generator that, on each pass through the data, produces a list of batches.
        """
        raise NotImplementedError

    def store(self, path):
        """Store the state of this module. Default is that there is no state, so nothing to store."""
        pass

    def load(self, path):
        """Load the state of this module. Default is that there is no state, so nothing to load."""
        pass


AnnotationType = TypeVar('AnnotationType')


class OnlineInputModule(InputModule, Generic[AnnotationType]):
    """InputModule that preprocesses datasets on the fly.

    It provides implementations for `create_batch()` and `__call__()` and
    introduces two abstract methods:
    - `preprocess()`: Converts a list of instances to annotations.
    - `create_batch()`: Converts a list of annotations to a batch.

    Both of these methods are parameterized by `AnnotationType`. In the simplest
    case, this could be a `dict`, but you could also define a separate class
    for your annotation, in order to get stronger typing.
    """

    def __init__(self, shared_resources: SharedResources, seed=None):
        self.shared_resources = shared_resources
        self._rng = random.Random(seed or random.randint(0, 9999))

    @abstractmethod
    def preprocess(self, questions: List[QASetting], answers: Optional[List[List[Answer]]] = None,
                   is_eval: bool = False) -> List[AnnotationType]:
        """Preprocesses a list of samples, returning a list of annotations.

        Batches of these annotation objects are then passed to the the `create_batch` method.

        Args:
            questions: The list of instances to preprocess
            answers: (Optional) answers associated with the instances
            is_eval: Whether this preprocessing is done for evaluation data

        Returns:
            List of annotations of the instances.
        """

        raise NotImplementedError

    @abstractmethod
    def create_batch(self, annotations: List[AnnotationType],
                     is_eval: bool, with_answers: bool) -> Mapping[TensorPort, np.ndarray]:
        """Creates a batch from a list of preprocessed questions.

        These are given by a list of annotations as returned by `preprocess_instance`.
        Args:
            annotations: a list of annotations to be included in the batch
            is_eval: whether the method is called for evaluation data
            with_answers: whether answers are included in the annotations

        Returns:
            A mapping from ports to numpy arrays.
        """

        raise NotImplementedError

    def _batch_questions(self, questions: List[Tuple[QASetting, List[Answer]]], batch_size, is_eval: bool):
        """Optionally shuffles and batches annotations.

        By default, all annotations are shuffled (if self.shuffle(is_eval) and
        then batched. Override this method if you want to customize the
        batching, e.g., to do stratified sampling, sampling with replacement,
        etc.

        Args:
            - annotations: List of annotations to shuffle & batch.
            - is_eval: Whether batches are generated for evaluation.

        Returns: Batch iterator
        """
        rng = self._rng if self._shuffle(is_eval) else None
        return shuffle_and_batch(questions, batch_size, rng)

    def _shuffle(self, is_eval: bool) -> bool:
        """Whether to shuffle the dataset in batch_annotations(). Default is noe is_eval."""
        return not is_eval

    def __call__(self, qa_settings: List[QASetting]) -> Mapping[TensorPort, np.ndarray]:
        """Preprocesses all qa_settings, returns a single batch with all instances."""

        annotations = self.preprocess(qa_settings, answers=None, is_eval=True)
        return self.create_batch(annotations, is_eval=True, with_answers=False)

    def batch_generator(self, dataset: List[Tuple[QASetting, List[Answer]]], batch_size: int, is_eval: bool) \
            -> Iterable[Mapping[TensorPort, np.ndarray]]:
        """Preprocesses all instances, batches & shuffles them and generates batches in dicts."""
        logger.info("OnlineInputModule pre-processes data on-the-fly in first epoch and caches results for subsequent "
                    "epochs! That means, first epoch might be slower.")
        # only cache training data on file
        use_cache = not is_eval and self.shared_resources.config.get('file_cache', False)
        if use_cache:
            cache_dir = os.path.join(os.environ.get('JACK_TEMP', tempfile.gettempdir()), 'cache')
            db = dc.Cache(cache_dir)
            db.reset('cull_limit', 0)
            logger.info("Caching temporary preprocessed data in %s. You can change cache dir using the"
                        " JACK_TEMP environment variable which defaults to /tmp/jack." % cache_dir)
        else:
            db = dict()
        preprocessed = set()
        def make_generator():
            running_idx = 0
            for i, batch in enumerate(self._batch_questions(dataset, batch_size, is_eval)):
                questions, answers = zip(*batch)
                if any(q.id not in preprocessed for q in questions):
                    annots = self.preprocess(questions, answers)
                    if questions[0].id is None:  # make sure there is an id, if not we set it here
                        for q in questions:
                            if q.id is None:
                                q.id = running_idx
                                running_idx += 1
                    for q, a in zip(questions, annots):
                        preprocessed.add(q.id)
                        db[q.id] = a
                else:
                    annots = [db[q.id] for q in questions]

                yield self.create_batch(annots, is_eval, True)

        return GeneratorWithRestart(make_generator)



# -*- coding: utf-8 -*-
import logging
from abc import abstractmethod
from typing import Mapping, List, Sequence

import numpy as np

from jack.core.tensorport import TensorPort

logger = logging.getLogger(__name__)


class ModelModule:
    """A model module defines the actual reader model by processing input tensors and producing output tensors.

    A model module encapsulates two computations (possibly overlapping): one which computes all
    predictions (to be processed by the output module) and another representing the loss(es) and potenially other
    training related outputs. It defines the expected input and output tensor shapes and types via its respective input
    and output pairs.
    """

    @abstractmethod
    def __call__(self, batch: Mapping[TensorPort, np.ndarray],
                 goal_ports: List[TensorPort] = None) -> Mapping[TensorPort, np.ndarray]:
        """Runs a batch and returns values/outputs for specified goal ports.
        Args:
            batch: mapping from ports to values
            goal_ports: optional output ports, defaults to output_ports of this module will be returned

        Returns:
            A mapping from goal ports to tensors.

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def output_ports(self) -> Sequence[TensorPort]:
        """Returns: Definition of the output ports of this module (predictions made by this model)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def input_ports(self) -> Sequence[TensorPort]:
        """Returns: Definition of the input ports."""
        raise NotImplementedError

    @property
    @abstractmethod
    def training_input_ports(self) -> Sequence[TensorPort]:
        """Returns: Definition of the input ports necessary to create the training output ports, i.e., they do not have
        to be provided during eval and they can include output ports of this module."""
        raise NotImplementedError

    @property
    @abstractmethod
    def training_output_ports(self) -> Sequence[TensorPort]:
        """Returns: Definition of the output ports provided during training for this module (usually just the loss)."""
        raise NotImplementedError

    @abstractmethod
    def setup(self, is_training=True, reuse=False):
        """Sets up the module."""
        raise NotImplementedError

    @abstractmethod
    def store(self, path):
        """Store the state of this module."""
        raise NotImplementedError

    @abstractmethod
    def load(self, path):
        """Load the state of this module."""
        raise NotImplementedError
